/*
 * Copyright (c) 2015-2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Wed Jul  6 2016
 */
#ifndef H_DYNAMIXEL_CODELS
#define H_DYNAMIXEL_CODELS

#include <errno.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>

#include "dynamixel_c_types.h"

struct dynamixel_channel_s {
  char path[1024];
  bool configured;
  bool present[256];
  int fd;

  uint8_t buf[128], r, w; /* read ring buffer */

  uint32_t header;
  uint8_t state;
  uint8_t id, error, p, param[256];
  uint16_t len, sum;
};

enum dx_cmd {
  DX_PING =		1,
  DX_READ =		2,
  DX_WRITE =		3,
  DX_REBOOT =		8,
  DX_SYNC_READ =	0x82,
  DX_SYNC_WRITE =	0x83
};

enum dx_mode {
  DX_TORQUE_MODE =		0,
  DX_VELOCITY_MODE =		1,
  DX_EXT_POSITION_MODE =	4,
  DX_TORQUE_POSITION_MODE =	5,
};

enum dx_ref {
  DX_MODEL_NUMBER =		0x00,
  DX_FIRMWARE_VERSION =		0x06,
  DX_ID =			0x07,
  DX_RETURN_DELAY_TIME =	0x09,
  DX_OPERATING_MODE =		0x0b,
  DX_SECONDARY_ID =		0x0b,
  DX_HOMING_OFFSET =		0x14,
  DX_TEMP_LIMIT =		0x1f,
  DX_MAX_VOLTAGE_LIMIT =	0x20,
  DX_MIN_VOLTAGE_LIMIT =	0x22,
  DX_PWM_LIMIT =		0x24,
  DX_CURRENT_LIMIT =		0x26,
  DX_ACCELERATION_LIMIT =	0x28,
  DX_VELOCITY_LIMIT =		0x2c,
  DX_TORQUE_ENABLE =		0x40,
  DX_LED =			0x41,
  DX_STATUS_RETURN_LEVEL =	0x44,
  DX_VELOCITY_I_GAIN =		0x4c,
  DX_VELOCITY_P_GAIN =		0x4e,
  DX_POSITION_D_GAIN =		0x50,
  DX_POSITION_I_GAIN =		0x52,
  DX_POSITION_P_GAIN =		0x54,
  DX_POSITION_KFF2_GAIN =	0x58,
  DX_POSITION_KFF1_GAIN =	0x5a,
  DX_GOAL_CURRENT =		0x66,
  DX_GOAL_VELOCITY =		0x68,
  DX_PROFILE_ACCELERATION =	0x6c,
  DX_PROFILE_VELOCITY =		0x70,
  DX_GOAL_POSITION =		0x74,
  DX_PRESENT_CURRENT =		0x7e,
};

extern const struct dx_modelid {
  int id; const char *name;
  dynamixel_resolution_s resolution;
} dx_models[];


static inline genom_event
dx_e_sys_error(const char *s, genom_context self)
{
  dynamixel_e_sys_detail d;
  size_t l = 0;

  d.code = errno;
  if (s) {
    strncpy(d.what, s, sizeof(d.what) - 3);
    l = strlen(s);
    strcpy(d.what + l, ": ");
    l += 2;
  }
  strerror_r(d.code, d.what + l, sizeof(d.what) - l);
  return dynamixel_e_sys(&d, self);
}

int	dx_open_tty(const char *device, speed_t baud);
int	dx_wait_msg(const struct dynamixel_channel_s *channel, int timeout);
int	dx_recv_msg(struct dynamixel_channel_s *chan, bool block);
int	dx_send_msg(const struct dynamixel_channel_s *chan, uint8_t id,
                enum dx_cmd cmd, ...);

void	dx_send_operating_mode(const dynamixel_channel_s *chan,
                unsigned int id, enum dx_mode mode);
void	dx_send_homing_offset(const dynamixel_channel_s *chan, unsigned int id,
                double offset, const dynamixel_resolution_s *resolution);
void	dx_recv_homing_offset(const uint8_t *buffer, double *offset,
                const dynamixel_resolution_s *resolution);
void	dx_send_limits(const dynamixel_channel_s *chan, unsigned int id,
                const struct dynamixel_limits_s *limits,
                const dynamixel_resolution_s *resolution);
void	dx_recv_limits(const uint8_t *buffer,
                struct dynamixel_limits_s *limits,
                const dynamixel_resolution_s *resolution);
void	dx_send_profile(const dynamixel_channel_s *chan, unsigned int id,
                const struct dynamixel_profile_s *profile,
                const dynamixel_resolution_s *resolution);
void	dx_recv_profile(const uint8_t *buffer,
                struct dynamixel_profile_s *profile,
                const dynamixel_resolution_s *resolution);
void	dx_send_position_pidff(const dynamixel_channel_s *chan, unsigned int id,
                const struct dynamixel_pidff_s *pidff);
void	dx_recv_position_pidff(const uint8_t *buffer,
                struct dynamixel_pidff_s *pidff);
void	dx_send_velocity_pi(const dynamixel_channel_s *chan, unsigned int id,
                const struct dynamixel_pi_s *pi);
void	dx_recv_velocity_pi(const uint8_t *buffer, struct dynamixel_pi_s *pi);

#endif /* H_DYNAMIXEL_CODELS */
