/*
 * Copyright (c) 2017 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Jul 24 2017
 */
#include "acdynamixel.h"

#include "dynamixel_c_types.h"
#include "codels.h"


/* --- operating mode ------------------------------------------------------ */

void
dx_send_operating_mode(const dynamixel_channel_s *chan, unsigned int id,
                       enum dx_mode mode)
{
  /* release motor & set led off */
  dx_send_msg(chan, id, DX_WRITE, 2, DX_TORQUE_ENABLE, 0, 0);

  dx_send_msg(chan, id, DX_WRITE, 1, DX_OPERATING_MODE, mode);

  /* enable motor & set led on */
  dx_send_msg(chan, id, DX_WRITE, 2, DX_TORQUE_ENABLE, 1, 1);
}


/* --- homing offset ------------------------------------------------------- */

void
dx_send_homing_offset(const dynamixel_channel_s *chan, unsigned int id,
                      double offset, const dynamixel_resolution_s *resolution)
{
  int32_t v32;

  v32 = offset / resolution->position;

  dx_send_msg(chan, id + 1, DX_WRITE, 1, DX_TORQUE_ENABLE, 0);
  dx_send_msg(chan, id + 1, DX_WRITE, 4, DX_HOMING_OFFSET,
              v32 & 0xff,
              (v32 >> 8) & 0xff,
              (v32 >> 16) & 0xff,
              (v32 >> 24) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 1, DX_TORQUE_ENABLE, 1);
}

void
dx_recv_homing_offset(const uint8_t *buffer, double *offset,
                      const dynamixel_resolution_s *resolution)
{
  int32_t v32;

  v32 = buffer[DX_HOMING_OFFSET];
  v32 |= ((uint32_t)buffer[DX_HOMING_OFFSET+1] << 8);
  v32 |= ((uint32_t)buffer[DX_HOMING_OFFSET+2] << 16);
  v32 |= ((int32_t)buffer[DX_HOMING_OFFSET+3] << 24);
  *offset = v32 * resolution->position;
}


/* --- limits -------------------------------------------------------------- */

void
dx_send_limits(const dynamixel_channel_s *chan, unsigned int id,
               const struct dynamixel_limits_s *limits,
               const dynamixel_resolution_s *resolution)
{
  uint8_t v8;
  uint16_t v16;
  uint32_t v32;

  dx_send_msg(chan, id + 1, DX_WRITE, 1, DX_TORQUE_ENABLE, 0);
  v8 = limits->temperature_max;
  dx_send_msg(chan, id + 1, DX_WRITE, 1, DX_TEMP_LIMIT, v8);

  v16 = limits->voltage_min * 10.;
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_MIN_VOLTAGE_LIMIT,
              v16 & 0xff, (v16 >> 8) & 0xff);

  v16 = limits->voltage_max * 10.;
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_MAX_VOLTAGE_LIMIT,
              v16 & 0xff, (v16 >> 8) & 0xff);

  v16 = limits->pwm_max * 885/100.;
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_PWM_LIMIT,
              v16 & 0xff, (v16 >> 8) & 0xff);

  v16 = limits->current_max / resolution->torque;
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_CURRENT_LIMIT,
              v16 & 0xff, (v16 >> 8) & 0xff);

  v32 = limits->acceleration_max / resolution->acceleration;
  dx_send_msg(chan, id + 1, DX_WRITE, 4, DX_ACCELERATION_LIMIT,
              v32 & 0xff, (v32 >> 8) & 0xff,
              (v32 >> 16) & 0xff, (v32 >> 24) & 0xff);

  v32 = limits->velocity_max / resolution->velocity;
  dx_send_msg(chan, id + 1, DX_WRITE, 4, DX_VELOCITY_LIMIT,
              v32 & 0xff, (v32 >> 8) & 0xff,
              (v32 >> 16) & 0xff, (v32 >> 24) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 1, DX_TORQUE_ENABLE, 1);
}

void
dx_recv_limits(const uint8_t *buffer, struct dynamixel_limits_s *limits,
               const dynamixel_resolution_s *resolution)
{
  uint16_t v16;
  uint32_t v32;

  limits->temperature_max = buffer[DX_TEMP_LIMIT];

  v16 = buffer[DX_MAX_VOLTAGE_LIMIT];
  v16 |= ((uint16_t)buffer[DX_MAX_VOLTAGE_LIMIT+1] << 8);
  limits->voltage_max = v16 / 10.;

  v16 = buffer[DX_MIN_VOLTAGE_LIMIT];
  v16 |= ((uint16_t)buffer[DX_MIN_VOLTAGE_LIMIT+1] << 8);
  limits->voltage_min = v16 / 10.;

  v16 = buffer[DX_PWM_LIMIT];
  v16 |= ((uint16_t)buffer[DX_PWM_LIMIT+1] << 8);
  limits->pwm_max = 100. * v16 / 885.;

  v16 = buffer[DX_CURRENT_LIMIT];
  v16 |= ((uint16_t)buffer[DX_CURRENT_LIMIT+1] << 8);
  limits->current_max = v16 * resolution->torque;

  v32 = buffer[DX_ACCELERATION_LIMIT];
  v32 |= ((uint32_t)buffer[DX_ACCELERATION_LIMIT+1] << 8);
  v32 |= ((uint32_t)buffer[DX_ACCELERATION_LIMIT+2] << 16);
  v32 |= ((uint32_t)buffer[DX_ACCELERATION_LIMIT+3] << 24);
  limits->acceleration_max = v32 * resolution->acceleration;

  v32 = buffer[DX_VELOCITY_LIMIT];
  v32 |= ((uint32_t)buffer[DX_VELOCITY_LIMIT+1] << 8);
  v32 |= ((uint32_t)buffer[DX_VELOCITY_LIMIT+2] << 16);
  v32 |= ((uint32_t)buffer[DX_VELOCITY_LIMIT+3] << 24);
  limits->velocity_max = v32 * resolution->velocity;
}


/* --- profile ------------------------------------------------------------- */

void
dx_send_profile(const dynamixel_channel_s *chan, unsigned int id,
                const struct dynamixel_profile_s *profile,
                const dynamixel_resolution_s *resolution)
{
  uint32_t v;

  v = profile->acceleration / resolution->acceleration;
  dx_send_msg(chan, id + 1, DX_WRITE, 4, DX_PROFILE_ACCELERATION,
              v & 0xff, (v >> 8) & 0xff, (v >> 16) & 0xff, (v >> 24) & 0xff);

  v = profile->velocity / resolution->velocity;
  dx_send_msg(chan, id + 1, DX_WRITE, 4, DX_PROFILE_VELOCITY,
              v & 0xff, (v >> 8) & 0xff, (v >> 16) & 0xff, (v >> 24) & 0xff);
}

void
dx_recv_profile(const uint8_t *buffer, struct dynamixel_profile_s *profile,
                const dynamixel_resolution_s *resolution)
{
  uint32_t v32;

  v32 = buffer[DX_PROFILE_ACCELERATION];
  v32 |= ((uint32_t)buffer[DX_PROFILE_ACCELERATION+1] << 8);
  v32 |= ((uint32_t)buffer[DX_PROFILE_ACCELERATION+2] << 16);
  v32 |= ((uint32_t)buffer[DX_PROFILE_ACCELERATION+3] << 24);
  profile->acceleration = v32 * resolution->acceleration;

  v32 = buffer[DX_PROFILE_VELOCITY];
  v32 |= ((uint32_t)buffer[DX_PROFILE_VELOCITY+1] << 8);
  v32 |= ((uint32_t)buffer[DX_PROFILE_VELOCITY+2] << 16);
  v32 |= ((uint32_t)buffer[DX_PROFILE_VELOCITY+3] << 24);
  profile->velocity = v32 * resolution->velocity;
}


/* --- position pidff ------------------------------------------------------ */

void
dx_send_position_pidff(const dynamixel_channel_s *chan, unsigned int id,
                       const struct dynamixel_pidff_s *pidff)
{
  uint16_t kP, kI, kD, kFF1, kFF2;

  kP = pidff->kP * 128;
  kI = pidff->kI * 65536;
  kD = pidff->kD * 16;
  kFF1 = pidff->kD * 4;
  kFF2 = pidff->kD * 4;

  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_POSITION_D_GAIN,
              kD & 0xff, (kD >> 8) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_POSITION_I_GAIN,
              kI & 0xff, (kI >> 8) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_POSITION_P_GAIN,
              kP & 0xff, (kP >> 8) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_POSITION_KFF2_GAIN,
              kFF2 & 0xff, (kFF2 >> 8) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_POSITION_KFF1_GAIN,
              kFF1 & 0xff, (kFF1 >> 8) & 0xff);
}

void
dx_recv_position_pidff(const uint8_t *buffer, struct dynamixel_pidff_s *pidff)
{
  uint16_t v16;

  v16 = buffer[DX_POSITION_D_GAIN];
  v16 |= ((uint16_t)buffer[DX_POSITION_D_GAIN + 1] << 8);
  pidff->kD = v16 / 16.;

  v16 = buffer[DX_POSITION_I_GAIN];
  v16 |= ((uint16_t)buffer[DX_POSITION_I_GAIN + 1] << 8);
  pidff->kI = v16 / 65536.;

  v16 = buffer[DX_POSITION_P_GAIN];
  v16 |= ((uint16_t)buffer[DX_POSITION_P_GAIN + 1] << 8);
  pidff->kP = v16 / 128.;

  v16 = buffer[DX_POSITION_KFF2_GAIN];
  v16 |= ((uint16_t)buffer[DX_POSITION_KFF2_GAIN + 1] << 8);
  pidff->kFF2 = v16 / 4.;

  v16 = buffer[DX_POSITION_KFF1_GAIN];
  v16 |= ((uint16_t)buffer[DX_POSITION_KFF1_GAIN + 1] << 8);
  pidff->kFF1 = v16 / 4.;
}


/* --- velocity pi --------------------------------------------------------- */

void
dx_send_velocity_pi(const dynamixel_channel_s *chan, unsigned int id,
                       const struct dynamixel_pi_s *pi)
{
  uint16_t kP, kI;

  kP = pi->kP * 128;
  kI = pi->kI * 65536;

  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_VELOCITY_I_GAIN,
              kI & 0xff, (kI >> 8) & 0xff);
  dx_send_msg(chan, id + 1, DX_WRITE, 2, DX_VELOCITY_I_GAIN,
              kP & 0xff, (kP >> 8) & 0xff);
}

void
dx_recv_velocity_pi(const uint8_t *buffer, struct dynamixel_pi_s *pi)
{
  uint16_t v16;

  v16 = buffer[DX_VELOCITY_I_GAIN];
  v16 |= ((uint16_t)buffer[DX_VELOCITY_I_GAIN + 1] << 8);
  pi->kI = v16 / 65536.;

  v16 = buffer[DX_VELOCITY_P_GAIN];
  v16 |= ((uint16_t)buffer[DX_VELOCITY_P_GAIN + 1] << 8);
  pi->kP = v16 / 128.;
}
